import jug.jug
import jug.task
from jug.task import Task
from jug.tests.utils import simple_execute
from jug.backends.dict_store import dict_store
from jug.options import parse
from .task_reset import task_reset_at_exit, task_reset
from .utils import find_test_jugfile
import pytest

import random
jug.jug.silent = True


@task_reset
def test_jug_execute_simple():
    N = 1024
    random.seed(232)
    A = [False for i in range(N)]
    def setAi(i):
        A[i] = True
    setall = [Task(setAi, i) for i in range(N)]
    store = dict_store()
    jug.task.Task.store = store
    simple_execute()
    assert False not in A
    assert max(store.counts.values()) < 4

@task_reset
def test_jug_execute_deps():
    N = 256
    random.seed(234)
    A = [False for i in range(N)]
    def setAi(i, other):
        A[i] = True
    idxs = list(range(N))
    random.shuffle(idxs)
    prev = None
    for idx in idxs:
        prev = Task(setAi, idx, prev)
    store = dict_store()
    jug.task.Task.store = store
    simple_execute()
    assert False not in A
    assert max(store.counts.values()) < 5

@pytest.mark.parametrize('jugfile',
      ['tasklet_simple.py',
      'tasklets.py',
      'barrier_mapreduce.py',
      'compound_nonsimple.py',
      'slice_task.py'])
def test_aggressive_unload(jugfile):
    from jug.jug import execution_loop
    from jug.task import alltasks

    options = parse(['execute'])
    options.aggressive_unload = True
    options.execute_target = None
    store, space = jug.jug.init(find_test_jugfile(jugfile), 'dict_store')
    execution_loop(alltasks, options)

@task_reset
def test_target_exact():
    from jug.jug import execution_loop
    from jug.task import alltasks
    options = parse(['execute'])
    options.jugfile = find_test_jugfile('simple.py')
    # Test if restricting to this target we skip the other tasks
    options.execute_target = "simple.double"

    store, space = jug.jug.init(options.jugfile, 'dict_store')
    execution_loop(alltasks, options)

    assert len(store.store) < len(alltasks)
    assert len(store.store) == 8

@task_reset
def test_target_wild():
    from jug.jug import execution_loop
    from jug.task import alltasks
    options = parse(['execute'])
    options.jugfile = find_test_jugfile('simple_multiple.py')
    # Test if restricting to this target we skip the other tasks
    options.execute_target = "simple_multiple.sum_"

    store, space = jug.jug.init(options.jugfile, 'dict_store')
    execution_loop(alltasks, options)

    assert len(store.store) < len(alltasks)
    assert len(store.store) == 16

@task_reset
def test_failed_task_keep_going():
    from jug.jug import execution_loop
    from jug.task import alltasks
    options = parse(['execute'])
    options.jugfile = find_test_jugfile('failing.py')
    # these 3 silence errors during execution and ensure jug isn't kept waiting
    options.execute_keep_going = True
    options.execute_nr_wait_cycles = 1
    options.execute_wait_cycle_time = 0
    # keep_failed ensures errored tasks are marked as failed
    options.execute_keep_failed = True

    options.execute_target = None
    store, space = jug.jug.init(options.jugfile, 'dict_store')
    # the failing.py jugfile has a total of 20 reachable tasks
    assert len(alltasks) == 20
    # Keep a copy of all tasks to check for failed tasks later
    alltasks_copy = alltasks[:]

    execution_loop(alltasks, options)

    # 14 results + 3 failures
    assert len(store.store) == 17
    # 3 tasks should remain in waiting state due to 3 failures upstream
    assert len(alltasks) == 3
    assert len([x for x in alltasks_copy if x.is_failed()]) == 3

@task_reset
def test_failed_task():
    from jug.jug import execution_loop
    from jug.task import alltasks
    from jug.tests.jugfiles.exceptions import FailingTask
    options = parse(['execute'])
    options.jugfile = find_test_jugfile('failing.py')
    # keep_failed ensures errored tasks are marked as failed
    options.execute_keep_failed = True
    options.execute_target = None

    store, space = jug.jug.init(options.jugfile, 'dict_store')
    # the failing.py jugfile has a total of 20 reachable tasks
    assert len(alltasks) == 20
    # Keep a copy of all tasks to check for failed tasks later
    alltasks_copy = alltasks[:]

    try:
        execution_loop(alltasks, options)
    except FailingTask:  # Using a custom exception to make sure we don't silence any errors
        pass

    # The third task fails so we get 2 results and 1 failed lock
    # NOTE: This might be incorrect if order of execution is not guaranteed
    assert len(store.store) == 3
    # 10 tasks could be run and were taken. Only the 10 waiting remain
    assert len(alltasks) == 10
    assert len([x for x in alltasks_copy if x.is_failed()]) == 1



import jug.jug
import jug.subcommands.execute
from .task_reset import task_reset_at_exit, task_reset
from .utils import simple_execute, find_test_jugfile

@task_reset
def test_no_load():
    jugfile = find_test_jugfile('no_load.py')
    store, space = jug.jug.init(jugfile, 'dict_store')
    x = space['x']
    assert not x.can_run()
    simple_execute()
    assert     x.can_run()



import inspect
from os import path
import pytest

@pytest.fixture(scope='function')
def tmp_file_store(tmpdir):
    from jug.backends.file_store import file_store
    from jug import task
    prev = task.Task.store
    yield file_store(str(tmpdir))
    task.Task.store = prev


def simple_execute():
    from jug.jug import execution_loop
    from jug.task import alltasks
    from jug.options import parse
    execution_loop(alltasks, parse(['execute']))

_jugdir = path.abspath(inspect.getfile(inspect.currentframe()))
_jugdir = path.join(path.dirname(_jugdir), 'jugfiles')

def find_test_jugfile(jugfile):
    return path.join(_jugdir, jugfile)




from time import sleep
from os import path

import jug.utils
import jug.task
from jug import value
from jug.backends.dict_store import dict_store

def test_util_timed_path(tmpdir):
    from jug.hash import hash_one
    jug.task.Task.store = dict_store()
    tmpdir = str(tmpdir)
    test_file = path.join(tmpdir, 'test_file')
    with open(test_file, 'wt') as out:
        out.write("Hello World")
    t0 = jug.utils.timed_path(test_file)
    t1 = jug.utils.timed_path(test_file)
    h0 = hash_one(t0)
    h1 = hash_one(t1)
    assert h0 == h1
    sleep(1.1)
    with open(test_file, 'wt') as out:
        out.write("Hello World")
    h1 = hash_one(t1)
    assert h0 != h1
    assert value(t0) == test_file
    assert value(t1) == test_file




import pytest

import jug.jug
from jug.tests.utils import simple_execute
from jug.task import describe
from .task_reset import task_reset_at_exit, task_reset
from .utils import find_test_jugfile


@task_reset
def test_describe():
    jugfile = find_test_jugfile('simple.py')
    store, space = jug.jug.init(jugfile, 'dict_store')
    simple_execute()
    t = space['vals'][0]
    desc = describe(t)
    assert len(desc['args']) == len(t.args)

@pytest.fixture
def __remove_files():
    yield
    from os import unlink
    from shutil import rmtree
    for f in ['x.pkl', 'x.meta.yaml', 'x.meta.json']:
        try:
            unlink(f)
        except:
            pass
    try:
        rmtree('testing_TO_DELETE.jugdata')
    except:
        pass

@task_reset
def test_describe_load(__remove_files):
    jugfile = find_test_jugfile('write_with_meta.py')
    store, space = jug.jug.init(jugfile, 'testing_TO_DELETE.jugdata')
    simple_execute()
    x = space['x']
    desc = describe(x)
    import json
    assert desc == json.load(open('x.meta.json'))
    import yaml
    assert desc == yaml.safe_load(open('x.meta.yaml'))



from jug import Task
from jug.task import iteratetask

def double(xs):
    return [x*2 for x in xs]

vals = [0,1,2]
t = Task(double, vals)
t0,t1,t2 = iteratetask(t, 3)

f2 = Task(double, [t0, t1])

f3 = Task(double, [2,3,4])




# -*- coding: utf-8 -*-
# Copyright (C) 2008-2020, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
file_store : file-system based data store & locks.
'''


import os
import sys
from os import path, mkdir
from os.path import dirname, exists
import errno
import logging
import tempfile
import shutil
from subprocess import Popen
from time import time

from .base import base_store, base_lock
from jug.backends.encode import encode_to, decode_from

def create_directories(dname):
    '''
    create_directories(dname)

    Recursively create directories.
    '''
    if dname.endswith('/'): dname = dname[:-1]
    head, tail = path.split(dname)
    if path.exists(dname): return
    if head: create_directories(head)
    try:
        mkdir(dname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def fsync_dir(fname):
    import errno
    parent = dirname(fname)
    try:
        fd = os.open(parent, os.O_RDONLY)
    except:
        # It seems that, on Windows and related platforms (cygwin...), you
        # cannot open a directory to get a file descriptor, so the call above
        # raises an error.
        import sys
        if not sys.platform.startswith('linux'):
            return
        else: # On Linux, we still want to check what's wrong
            raise
    try:
        os.fsync(fd)
    except OSError as err:
        if err.errno != errno.EINVAL:
            raise
    finally:
        os.close(fd)


class file_store(base_store):
    def __init__(self, dname, compress_numpy=False):
        '''
        file_store(dname)

        Recursively create directories.
        '''
        if dname.endswith('/'): dname = dname[:-1]
        self.jugdir = dname
        self.compress_numpy = compress_numpy

    def __repr__(self):
        return 'file_store({})'.format(self.jugdir)
    __str__ = __repr__

    def create(self):
        '''
        Recursively create directories.
        '''
        create_directories(self.jugdir)
        create_directories(self.tempdir())

    def _maybe_create(self):
        '''
        Calls self.create() the first time it is called; then becomes a no-op.
        '''
        self.create()
        self._maybe_create = (lambda : None)

    def tempdir(self):
        return path.join(self.jugdir, 'tempfiles')

    def _getfname(self, name):
        if type(name) != str:
            name = name.decode('utf-8')
        return path.join(self.jugdir, name[:2], name[2:])


    def dump(self, object, name):
        '''
        store.dump(object, name)

        Performs roughly the same as

        pickle.dump(object, open(name,'w'))

        but does it in a way that is guaranteed to be atomic even over NFS and
        using compression on the disk for faster access.
        '''
        self._maybe_create()
        name = self._getfname(name)
        create_directories(dirname(name))
        fd, fname = tempfile.mkstemp('.jugtmp', 'jugtemp', self.tempdir())
        output = os.fdopen(fd, 'wb')
        try:
            import numpy as np
            if not self.compress_numpy and type(object) == np.ndarray:
                np.lib.format.write_array(output, object)
                output.flush()
                os.fsync(output.fileno())
                output.close()
                fsync_dir(fname)
                os.rename(fname, name)
                return
        except ImportError:
            pass
        except OSError:
            pass
        except ValueError:
            pass

        encode_to(object, output)
        output.flush()
        os.fsync(output.fileno())
        output.close()

        # Rename is atomic even over NFS.
        fsync_dir(fname)
        os.rename(fname, name)

    def list(self):
        '''
        keys = store.list()

        Returns a list of all the keys in the store
        '''
        if not exists(self.jugdir):
            return []

        keys = []
        for d in os.listdir(self.jugdir):
            if len(d) == 2:
                for f in os.listdir(path.join(self.jugdir, d)):
                    keys.append((d+f).encode('ascii'))
        return keys


    def listlocks(self):
        '''
        keys = store.listlocks()

        Returns a list of all the locks in the store

        This is an unsafe function as the results may be outdated by the time
        the function returns.
        '''
        if not exists(path.join(self.jugdir, 'locks')):
            return []

        keys = []
        for k in os.listdir(path.join(self.jugdir, 'locks')):
            keys.append(k[:-len('.lock')].encode('ascii'))
        return keys


    def can_load(self, name):
        '''
        can = store.can_load(name)
        '''
        fname = self._getfname(name)
        return exists(fname)


    def load(self, name):
        '''
        obj = store.load(name)

        Loads the objects. Equivalent to pickle.load(), but a bit smarter at
        times.

        Parameters
        ----------
        name : str
            Key to use

        Returns
        -------
        obj : any
            The object that was saved under ``name``
        '''
        fname = self._getfname(name)
        with open(fname, 'rb') as ifile:
            try:
                import numpy as np
                return np.lib.format.read_array(ifile)
            except ValueError:
                ifile.seek(0)
            except ImportError:
                pass
            return decode_from(ifile)

    def remove(self, name):
        '''
        was_removed = store.remove(name)

        Remove the entry associated with name.

        Returns whether any entry was actually removed.
        '''
        try:
            fname = self._getfname(name)
            os.unlink(fname)
            return True
        except OSError:
            return False

    def cleanup(self, active, keeplocks=False):
        '''
        nr_removed = store.cleanup(active, keeplocks)

        Implement 'cleanup' command

        Parameters
        ----------
        active : sequence
            files *not to remove*
        keeplocks : boolean
            whether to preserve or remove locks

        Returns
        -------
        nr_removed : integer
            number of removed files
        '''
        active = frozenset(self._getfname(t.hash()) for t in active)
        removed = 0
        for dir,_,fs in os.walk(self.jugdir):
            if keeplocks and dir == "locks":
                continue
            for f in fs:
                f = path.join(dir, f)
                if f not in active:
                    os.unlink(f)
                    removed += 1
        return removed

    def remove_locks(self):
        '''
        removed = store.remove_locks()

        Remove all locks

        Returns
        -------
        removed : int
            Number of locks removed
        '''

        lockdir = path.join(self.jugdir, 'locks')
        if not exists(lockdir): return 0

        removed = 0
        for f in os.listdir(lockdir):
            os.unlink(path.join(lockdir, f))
            removed += 1
        return removed

    def getlock(self, name):
        '''
        lock = store.getlock(name)

        Retrieve a lock object associated with ``name``.

        Parameters
        ----------
        name : str
            Key

        Returns
        -------
        lock : Lock object
            This is a file_lock object
        '''
        return file_based_lock(self.jugdir, name)

    def close(self):
        '''
        store.close()

        Has no effect on file based stores.
        '''
        pass

    def metadata(self, t):
        '''
        meta = store.metadata(t)

        Retrieves information on the state of the computation

        Parameters
        ----------
        t : Task
            A Task object

        Returns
        -------
        meta : dict
            Dictionary describing the state of the computation
        '''
        from os import stat, path
        from time import ctime
        fname = self._getfname(t.hash())
        if path.exists(fname):
            st = stat(fname)
            return {
                'computed': True,
                'completed': ctime(st.st_mtime),
            }
        return {
                'computed': False
        }


    @staticmethod
    def remove_store(jugdir):
        '''
        file_store.remove_store(jugdir)

        Removes from disk all the files associated with this jugdir.
        '''
        if path.exists(jugdir):
            shutil.rmtree(jugdir)


class file_based_lock(base_lock):
    '''
    file_based_lock: File-system based locks

    Functions:
    ----------

    - get(): acquire the lock
    - release(): release the lock
    - is_locked(): check lock state
    '''
    # We use epoch + 1 second to mark a task as failed
    # More at https://github.com/luispedro/jug/issues/55
    _FAILED_TIMESTAMP = (1, 1)  # atime, mtime

    def __init__(self, jugdir, name):
        import six
        if type(name) != six.text_type:
            name = six.text_type(name, 'utf-8')
        self.fullname = path.join(jugdir, 'locks', '{0}.lock'.format(name))

    def get(self):
        '''
        lock.get()

        Create a lock for name in an NFS compatible way.

        Parameters
        ----------
        None

        Returns
        -------
        locked : bool
            Whether the lock was created
        '''
        if exists(self.fullname): return False
        create_directories(path.dirname(self.fullname))
        try:
            import socket
            from datetime import datetime
            fd = os.open(self.fullname, os.O_RDWR|os.O_CREAT|os.O_EXCL)
            F = os.fdopen(fd, 'w')
            F.write('PID {0} on HOSTNAME {1}\n'.format(os.getpid(), socket.gethostname()))
            F.write('Lock created on {0}\n'.format(datetime.now().strftime('%Y-%m-%d (%Hh%M.%S)')))
            F.close()
            return True
        except OSError:
            return False

    def release(self):
        '''
        lock.release()

        Removes lock
        '''
        try:
            os.unlink(self.fullname)
        except OSError:
            pass

    def is_locked(self):
        '''
        locked = lock.is_locked()

        Returns whether a lock exists for name. Note that the answer can
        be invalid by the time this function returns. Only by trying to
        acquire the lock can you avoid race-conditions. See the get() function.
        '''
        return path.exists(self.fullname)

    def fail(self):
        '''
        lock.fail()

        Mark a task as failed.
        Has no effect if the task isn't locked
        '''
        try:
            os.utime(self.fullname, self._FAILED_TIMESTAMP)
        except OSError:
            return False
        else:
            return True

    def is_failed(self):
        '''
        failed = lock.is_failed()

        Returns whether this task is marked as failed.

        This code is not race-condition free. It may happen that by the time
        this function returns, the failed lock has been released.
        '''
        if self.is_locked():
            try:
                t = os.stat(self.fullname)
            except OSError:
                pass
            else:
                if t.st_mtime == self._FAILED_TIMESTAMP[0]:
                    return True

        return False


class file_keepalive_store(file_store):
    def __repr__(self):
        return 'file_keepalive_store({})'.format(self.jugdir)
    __str__ = __repr__

    def getlock(self, name):
        '''
        lock = store.getlock(name)

        Retrieve a lock object associated with ``name``.

        Parameters
        ----------
        name : str
            Key

        Returns
        -------
        lock : Lock object
            This is a file_lock object
        '''
        return file_keepalive_based_lock(self.jugdir, name)


class file_keepalive_based_lock(file_based_lock):
    '''
    file_keepalive_based_lock: File-system based locks

    Works much like file_based_lock but implements a mechanism to keep locks
    alive during execution, making it possible to recognize crashed/killed jobs
    as failed.

    Functions:
    ----------

    - get(): acquire the lock
    - release(): release the lock
    - is_locked(): check lock state
    '''
    def __init__(self, *args, **kwargs):
        self.monitor = None
        super(file_keepalive_based_lock, self).__init__(*args, **kwargs)

    def start_monitor(self):
        """Start side-kick process that ensures locks are refreshed while
        main process is active on this task
        """
        # No need to be gentle. Killing it straight away is safe.
        self.monitor = Popen([sys.executable, "-m", "jug.backends.file_keepalive_monitor", self.fullname])

    def stop_monitor(self):
        """Stop side-kick process that ensures locks are refreshed while
        main process is active on this lock
        """
        if self.monitor is None:
            return

        try:
            self.monitor.kill()
        except OSError as e:
            logging.warning('keepalive process failed to die with %s' % e)

        # subprocess module implements a cleanup mechanism that kicks in when
        # a Popen instance is deleted. This prevents leaking subprocesses and
        # avoids having to explicitly block and .wait() for a task to finish
        del self.monitor
        self.monitor = None

    def get(self):
        '''
        lock.get()

        Create a lock for name in an NFS compatible way.
        And start a lock monitor job

        Parameters
        ----------
        None

        Returns
        -------
        locked : bool
            Whether the lock was created
        '''
        acquired = super(file_keepalive_based_lock, self).get()

        if acquired:
            self.start_monitor()

        return acquired

    def release(self):
        '''
        lock.release()

        Removes lock
        And stops lock monitor job
        '''
        self.stop_monitor()
        return super(file_keepalive_based_lock, self).release()

    def fail(self):
        '''
        lock.fail()

        Mark a task as failed.
        Has no effect if the task isn't locked other than stopping the monitoring
        process if it exists
        '''
        self.stop_monitor()
        return super(file_keepalive_based_lock, self).fail()

    def is_failed(self):
        '''
        failed = lock.is_failed()

        Returns whether this task has a lock in failed state.

        A lock in failed state is one that exists and was last modified more
        than 30 minutes ago.

        This code is not race-condition free. It may happen that by the time
        this function returns, the failed lock has been released.
        '''
        failed_lock = time() - (30 * 60)  # 30 minutes old lock is a dead lock

        if self.is_locked():
            try:
                t = os.stat(self.fullname)
            except OSError:
                pass
            else:
                if t.st_mtime <= failed_lock:
                    return True

        return False



import pytest
import jug.task
from jug.backends.dict_store import dict_store
from jug.hooks import reset_all_hooks

@pytest.fixture(scope='function')
def task_reset_at_exit():
    jug.task.Task.store = dict_store()
    while jug.task.alltasks:
        jug.task.alltasks.pop()

    yield

    jug.task.Task.store = None
    while jug.task.alltasks:
        jug.task.alltasks.pop()
    reset_all_hooks()

def task_reset(f):
    return pytest.mark.usefixtures('task_reset_at_exit')(f)



from jug import TaskGenerator
from jug.tests.jugfiles.exceptions import FailingTask

@TaskGenerator
def some_fail(x):
    if x in (2, 5, 8):
        raise FailingTask

    return x

@TaskGenerator
def plus1(x):
    return x + 1

vals = list(map(some_fail, list(range(10))))
vals = list(map(plus1, vals))

# This jugfile has 20 tasks of which 10 are executable and 10 are waiting.
# If the first 10, 3 will fail.
# With full execution this should result in 14 complete tasks, 3 failed and 3 waiting



import os
import jug.backends.redis_store
import jug.backends.file_store
import jug.backends.dict_store
from jug.backends.redis_store import redis
import pytest
import six

if not os.getenv('TEST_REDIS'):
    redis = None

try:
    redisConnectionError = redis.ConnectionError
except:
    redisConnectionError = SystemError

@pytest.fixture(scope='function', params=['file', 'dict', 'redis'])
def store(tmpdir, request):
    if request.param == 'file':
        tmpdir = str(tmpdir)
        yield jug.backends.file_store.file_store(tmpdir)
        jug.backends.file_store.file_store.remove_store(tmpdir)
    elif request.param == 'dict':
        yield jug.backends.dict_store.dict_store()
    elif request.param == 'redis':
        if redis is None:
            pytest.skip()
        try:
            st = jug.redis_store.redis_store('redis:')
            yield st
            st.close()
        except redisConnectionError:
            pytest.skip()

def test_load_get(store):
    assert len(list(store.list())) == 0
    key = six.b('jugisbestthingever')
    assert not store.can_load(key)
    object = list(range(232))
    store.dump(object, key)
    assert store.can_load(key)
    assert store.load(key) == object

    flist = list(store.list())
    assert len(flist) == 1
    assert flist[0] == key

    store.remove(key)
    assert not store.can_load(key)

def test_lock(store):
    assert len(list(store.listlocks())) == 0
    key = six.b('jugisbestthingever')
    lock = store.getlock(key)
    assert not lock.is_locked()
    assert lock.get()
    assert not lock.get()
    lock2 = store.getlock(key)
    assert not lock2.get()
    assert len(list(store.listlocks())) == 1
    lock.release()
    assert lock2.get()
    lock2.release()

def test_lock_remove(store):
    assert len(list(store.listlocks())) == 0
    key = six.b('jugisbestthingever')
    lock = store.getlock(key)
    assert not lock.is_locked()
    assert lock.get()
    assert not lock.get()
    assert len(list(store.listlocks())) == 1
    store.remove_locks()
    assert len(list(store.listlocks())) == 0

def test_lock_fail(store):
    assert len(list(store.listlocks())) == 0
    key = six.b('jugisbestthingever')
    lock = store.getlock(key)
    assert not lock.is_locked()
    assert lock.get()
    assert not lock.get()
    lock.fail()
    assert lock.is_failed()
    assert len(list(store.listlocks())) == 1
    store.remove_locks()
    assert not lock.is_failed()
    assert len(list(store.listlocks())) == 0

def test_lock_fail_other(store):
    # is_failed should return True even if we can't acquire the lock
    assert len(list(store.listlocks())) == 0
    key = six.b('jugisbestthingever')
    lock1 = store.getlock(key)
    lock2 = store.getlock(key)
    assert not lock1.is_locked()
    assert not lock2.is_locked()
    assert lock1.get()
    assert not lock2.get()
    assert not lock1.is_failed()
    assert not lock2.is_failed()
    lock1.fail()
    assert lock2.is_failed()
    assert len(list(store.listlocks())) == 1
    store.remove_locks()
    assert not lock1.is_failed()
    assert not lock2.is_failed()
    assert len(list(store.listlocks())) == 0

def test_numpy_array(tmpdir):
    try:
        import numpy as np
    except ImportError:
        pytest.skip()
    store = jug.backends.file_store.file_store(str(tmpdir))
    arr = np.arange(100) % 17
    arr = arr.reshape((10,10))
    key = 'mykey'
    store.dump(arr, key)
    arr2 = store.load(key)
    assert np.all(arr2 == arr)
    store.remove(key)
    store.close()

def test_numpy_array_no_compress(tmpdir):
    try:
        import numpy as np
    except ImportError:
        pytest.skip()
    store = jug.backends.file_store.file_store(str(tmpdir), compress_numpy=False)
    arr = np.arange(100) % 17
    arr = arr.reshape((10,10))
    key = 'mykey'
    store.dump(arr, key)
    arr2 = store.load(key)
    assert np.all(arr2 == arr)
    store.remove(key)
    store.close()



import jug
from jug.io import NoLoad
import random

@jug.TaskGenerator
def gauss():
    return random.gauss(0, 1)

@jug.TaskGenerator
def sum_partials(ts):
    total = 0.0

    for t in ts:
        total+= jug.value(t.base)
    return total

@jug.TaskGenerator
def double(x):
    return 2*x

map_ts = [gauss() for i in range(10)]
x = sum_partials([NoLoad(t) for t in map_ts])
final = double(x)



from jug import Task, TaskGenerator
from jug import mapreduce



def double(x):
    return 2*x

def double2(x):
    return 2*x

@TaskGenerator
def do_sum(partials):
    return sum(partials)

data = list(range(16))

partial = [Task(double, x) for x in data]
repartial = mapreduce.map(double2, partial, map_step=2)

final = do_sum(repartial)




from jug.backends.file_store import file_store, file_based_lock, file_keepalive_based_lock
from jug.backends.dict_store import dict_store
from jug.tests.utils import tmp_file_store
from .task_reset import task_reset_at_exit, task_reset
from jug.backends import memoize_store
from jug import Task
from time import sleep


def test_twice(tmpdir):
    lock = file_based_lock(str(tmpdir), 'foo')
    assert lock.get()
    assert not lock.get()
    lock.release()

    assert lock.get()
    assert not lock.get()
    lock.release()

def test_twolocks(tmpdir):
    foo = file_based_lock(str(tmpdir), 'foo')
    bar = file_based_lock(str(tmpdir), 'bar')
    assert foo.get()
    assert bar.get()
    assert not foo.get()
    assert not bar.get()
    foo.release()
    bar.release()


def test_fail_and_lock(tmpdir):
    lock = file_based_lock(str(tmpdir), 'foo')
    assert not lock.is_failed()
    assert not lock.is_locked()

    assert not lock.fail()
    assert not lock.is_failed()
    assert not lock.is_locked()

    assert lock.get()
    assert not lock.is_failed()
    assert lock.is_locked()

    assert lock.fail()
    assert lock.is_failed()
    assert lock.is_locked()

    assert lock.fail()
    assert lock.is_failed()
    assert lock.is_locked()

    assert not lock.get()
    assert lock.is_failed()
    assert lock.is_locked()

    lock.release()
    assert not lock.is_failed()
    assert not lock.is_locked()


def double(x):
    return x*2

@task_reset
def test_memoize_lock(tmp_file_store):

    t = Task(double, 2)
    assert t.lock()

    Task.store = memoize_store(Task.store, list_base=True)
    assert t.is_locked()
    t2 = Task(double, 2)
    assert t2.is_locked()

def test_lock_bytes(tmp_file_store):
    store = tmp_file_store
    lock = store.getlock('foo')
    lock2 = store.getlock(b'foo')
    assert lock.fullname == lock2.fullname

def test_lock_bytes2():
    store = dict_store()
    lock = store.getlock('foo')
    lock2 = store.getlock(b'foo')
    lock.get()
    assert lock2.is_locked()

def test_lock_keepalive(tmpdir):
    lock = file_keepalive_based_lock(str(tmpdir), 'foo')
    assert lock.monitor is None
    assert lock.get()
    assert lock.monitor.poll() is None
    p = lock.monitor
    assert not lock.get()
    assert p == lock.monitor, "A new process was started and shouldn't"
    assert lock.monitor.poll() is None
    lock.release()

    # Give max 5 secs for subprocess to return an exitcode
    for i in range(25):
        ret = p.poll()
        if ret is None:
            sleep(.2)
        else:
            break

    assert ret == -9  # SIGKILL

    assert lock.get()
    assert p != lock.monitor, "A new process should have been started but wasn't"
    assert lock.monitor.poll() is None
    p = lock.monitor
    assert not lock.get()
    assert lock.monitor.poll() is None
    assert p == lock.monitor, "A new process was started and shouldn't"
    lock.release()

    # Give max 5 secs for subprocess to return an exitcode
    for i in range(25):
        ret = p.poll()
        if ret is None:
            sleep(.2)
        else:
            break

    assert ret == -9  # SIGKILL



from jug import TaskGenerator

@TaskGenerator
def double(x):
    return 2*x

@TaskGenerator
def create_result(r, oname):
    with open(oname, 'wt') as out:
        out.write('Result is {}\n'.format(r))

two = double(1)
four = double(2)
eight = double(four)
sixteen = double(eight)

create_result(sixteen, 'test-result.txt')






# -*- coding: utf-8 -*-
# Copyright (C) 2008-2017, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
file_store_monitor : part of file_store to keep locks alive during execution
'''

from os import utime, getppid, kill
from sys import argv
from time import sleep


def parent_gone_or_changed(pid):
    current_parent = getppid()
    if current_parent != pid or current_parent == 1:
        # A different parent means parent died and/or we got disowned, neither
        # of which should happen in normal conditions
        # If PID 1 we are running under init and that also means we got disowned
        return True

    # Test if parent is still alive
    try:
        kill(pid, 0)
    except OSError:
        return True
    else:
        return False


def main():
    lock = argv[1]
    parent = getppid()
    # 5 * 60 = once every 5 minutes
    counter = counter_start = 60

    while True:
        # self check every 5 seconds
        sleep(5)

        # We die if our parent went away
        if parent_gone_or_changed(parent):
            break

        # but only update the lock once every 5 minutes for IO reasons
        counter -= 1

        if counter <= 0:
            counter = counter_start
            try:
                utime(lock, None)
            except OSError:
                # Lock is no longer available
                break


if __name__ == '__main__':
    main()



from jug import mapreduce, TaskGenerator

@TaskGenerator
def double(x):
    return 2*x

@TaskGenerator
def sum_all(xs):
    return sum(xs)

vs = list(range(16))
v2s = mapreduce.map(double, vs)

s = sum_all(v2s)
v4s = mapreduce.map(double, v2s)



import pytest

from jug.hooks import register_hook, register_hook_once, reset_all_hooks, jug_hook

A = []
def inc_A():
    A.append(0)


@pytest.fixture(scope='function')
def reset_hooks_at_exit():
    yield
    reset_all_hooks()
    del A[:] # A.clear() is not available in Python 2.x

def test_basic(reset_hooks_at_exit):
    assert len(A) == 0
    register_hook('execute.task-executed1', inc_A)
    jug_hook('execute.task-executed1')
    assert len(A) == 1

def test_register_twice(reset_hooks_at_exit):
    register_hook('execute.task-executed1', inc_A)
    register_hook('execute.task-executed1', inc_A)
    jug_hook('execute.task-executed1')
    assert len(A) == 2

def test_reset_all_hooks(reset_hooks_at_exit):
    register_hook('execute.task-executed1', inc_A)
    jug_hook('execute.task-executed1')
    assert len(A) == 1
    reset_all_hooks()
    jug_hook('execute.task-executed1')
    assert len(A) == 1

def test_register_once(reset_hooks_at_exit):
    assert len(A) == 0
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    assert len(A) == 0
    jug_hook('execute.task-executed1')
    assert len(A) == 1

def test_bad_name():
    with pytest.raises(ValueError):
        register_hook('my_bad_hook', inc_A)




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2020, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.


def _get_debugger():
    '''
    Get a debugger object
    '''
    # The code below is a complex attempt to load IPython
    # debugger which works with multiple versions of IPython.
    #
    # Unfortunately, their API kept changing prior to version 1.0.
    try:
        import IPython
        try:
            import IPython.core.debugger
            try:
                from IPython.terminal.ipapp import load_default_config
                config = load_default_config()
                colors = config.TerminalInteractiveShell.colors
            except:
                import IPython.core.ipapi
                ip = IPython.core.ipapi.get()
                colors = ip.colors
            try:
                return IPython.core.debugger.Pdb(colors.get_value(initial='Linux'))
            except AttributeError:
                return IPython.core.debugger.Pdb(colors)
        except ImportError:
            #Fallback to older version of IPython API
            import IPython.ipapi
            import IPython.Debugger
            shell = IPython.Shell.IPShell(argv=[''])
            ip = IPython.ipapi.get()
            return IPython.Debugger.Pdb(ip.options.colors)
    except ImportError:
        #Fallback to standard debugger
        import pdb
        return pdb.Pdb()


def debug_exception():
    '''
    Start a debugger on the exception currently being handled
    '''
    import sys

    _,_, tb = sys.exc_info()
    debugger = _get_debugger()
    debugger.reset()
    debugger.interaction(None, tb)
